import platform

from pyqalx.config import UserConfig, Config
from pyqalx.core.adapters import (
    QalxBlueprint,
    QalxBot,
    QalxGroup,
    QalxItem,
    QalxQueue,
    QalxSet,
    QalxWorker,
    QalxFactory,
    QalxNotification,
)
from pyqalx.core.log import configure_logging, session_logging_suffix
from pyqalx.core.registry import Registry
from pyqalx.core.tags import Tags
from pyqalx.transport import PyQalxAPI


class QalxSession:
    """The session that any interaction with the API will use."""

    def __init__(
        self,
        profile_name=Config.default,
        config_class=None,
        skip_ini=False,
        rest_api_class=None,
        project_name=None,
    ):
        """
        :param profile_name:    profile name to get from `config_class`
                                (default=Config.default)
        :type profile_name: str
        :param config_class:  The class for the config that will be used for
                                this session
        :type config_class: ~pyqalx.config.configs.Config
        :param skip_ini: Should loading the config from the inifile be skipped?
        :type skip_ini: bool
        :param rest_api_class: The class to use for the rest api
        :type rest_api_class: ~pyqalx.transport.api.PyQalxAPI
        :param project_name: An optional name of a project. This will be used
                              to tag all data that is created/modified with this session
        :type project_name: str
        """

        if config_class is None:
            config_class = UserConfig
        self.config = config_class()
        self.profile_name = profile_name
        if not skip_ini:
            self.config.from_inifile(profile_name)
        # Configure the log for this session.  This cannot be set on an
        # instance attribute due to pickling issues with windows.  However,
        # it needs to be configured as early as possibly after config
        # initialisation so that if there is an unhandled exception this can be
        # logged properly
        self.configure_logging()

        if rest_api_class is None:
            rest_api_class = PyQalxAPI
        self.rest_api = rest_api_class(self)
        self._registry = Registry()
        self._registry._register_defaults()
        self.tags = Tags([], rest_api=self.rest_api)
        if project_name:
            self.tags.add(name="project_name", value=project_name)

    def configure_logging(self):
        """
        Configures the log for this session.  A log is namespaced for the
        config type and the profile name to allow a user to have multiple
        session instances with different configs/profiles logging to different
        places.

        :return: An instance of `logging.getLogger`
        """
        # Ensures that the log directories are created if they don't already
        # exist
        self.config._create_log_directories()
        return configure_logging(
            f"{session_logging_suffix(self.config)}.{self.profile_name}",
            self.config.get("LOG_FILE_PATH"),
            self.config.get("LOGGING_LEVEL"),
        )

    @property
    def log(self):
        """
        Returns an instance of `logging.getLogger` for this QalxSession instance
        This cannot be set as an instance attribute due to pickling issues on
        windows.  Will log depending on the `LOGGING_LEVEL` config value.

        Usage:

        >>> QalxSession().log.debug("A debug message")
        >>> QalxSession().log.info("An info message")
        >>> QalxSession().log.warning("A warning message")
        >>> QalxSession().log.error("An error message")

        :returns: `logging.getLogger`
        """
        return self.configure_logging()

    def __setattr__(self, key, value):
        if key == "tags" and hasattr(self, "tags"):
            # The user is completely replacing the tags.
            # Attempt to validate them before replacing them.
            self.tags._validate(value)
        return super(QalxSession, self).__setattr__(key, value)

    def _update_config(self, config):
        """
        Method to use if the config needs to be updated after the session has
        been created.
        Also updates the rest_api with the updated config

        :param config:A config dict
        :type config: dict
        :return:
        """
        self.config.update(config)
        self.rest_api = self.rest_api.__class__(self)

    def register(self, cls):
        """
        Registers the given :class:`~pyqalx.core.entities.entity.QalxEntity`
        subclass with the session.   See entities docs for more info

        :param cls: subclass of QalxEntity.
        :type cls: pyqalx.core.entities.entity.QalxEntity
        """
        self._registry.register(cls)

    def unregister(self, cls):
        """
        Unregisters the given :class:`~pyqalx.core.entities.entity.QalxEntity`
        subclass from the session.   See entities docs for more info

        :param cls: subclass of QalxEntity.
        :type cls: pyqalx.core.entities.entity.QalxEntity
        """
        self._registry.unregister(cls)

    @property
    def _host_info(self):
        return {
            "node": platform.node(),
            "platform": platform.platform(),
            "ip": self.rest_api._get_external_ip(),
        }

    @property
    def item(self):
        """
        returns a :class:`~pyqalx.core.adapters.item.QalxItem` adapter for
        this session

        :return: Instance of :class:`~pyqalx.core.adapters.item.QalxItem`
        """
        return QalxItem(self)

    @property
    def set(self):
        """
        returns a :class:`~pyqalx.core.adapters.set.QalxSet` adapter for
        this session

        :return: Instance of :class:`~pyqalx.core.adapters.set.QalxSet`
        """
        return QalxSet(self)

    @property
    def group(self):
        """
        returns a :class:`~pyqalx.core.adapters.group.QalxGroup` adapter for
        this session

        :return: Instance of :class:`~pyqalx.core.adapters.group.QalxGroup`
        """
        return QalxGroup(self)

    @property
    def queue(self):
        """
        returns a :class:`~pyqalx.core.adapters.queue.QalxQueue` adapter for
        this session

        :return: Instance of :class:`~pyqalx.core.adapters.queue.QalxQueue`
        """
        return QalxQueue(self)

    @property
    def bot(self):
        """
        returns a :class:`~pyqalx.core.adapters.bot.QalxBot` adapter for
        this session

        :return: Instance of :class:`~pyqalx.core.adapters.bot.QalxBot`
        """
        return QalxBot(self)

    @property
    def worker(self):
        """
        returns a :class:`~pyqalx.core.adapters.worker.QalxWorker` adapter
        for this session

        :return: Instance of :class:`~pyqalx.core.adapters.worker.QalxWorker`
        """
        return QalxWorker(self)

    @property
    def factory(self):
        """
        returns a :class:`~pyqalx.core.adapters.factory.QalxFactory` adapter
        for this session

        :return: Instance of :class:`~pyqalx.core.adapters.factory.QalxFactory`
        """
        return QalxFactory(self)

    @property
    def blueprint(self):
        """
        returns a :class:`~pyqalx.core.adapters.blueprint.QalxBlueprint`
        adapter for this session

        :return: Instance of :class:`~pyqalx.core.adapters.blueprint.QalxBlueprint`
        """
        return QalxBlueprint(self)

    @property
    def notification(self):
        """
        returns a :class:`~pyqalx.core.adapters.notification.QalxNotification`
        adapter for this session

        :return: Instance of
                :class:`~pyqalx.core.adapters.notification.QalxNotification`
        """
        return QalxNotification(self)



"""
Sessions
~~~~~~~~~~~~~~~~~~~

The best way to interface between entities and qalx is with the `session` object.
This has a series of helper methods to interface with the API using entity
specific `adapters`

.. _qalxsession:

.. automodule:: pyqalx.core.session
    :members:

Adapters
~~~~~~~~

Adapters are how a :class:`QalxSession` interfaces with the API.  These won't typically
get accessed directly - instead the helper properties on the
:class:`QalxSession` should be used.

.. automodule:: pyqalx.core.adapters.blueprint
     :members:

.. automodule:: pyqalx.core.adapters.bot
     :members:

.. automodule:: pyqalx.core.adapters.factory
     :members:

.. automodule:: pyqalx.core.adapters.group
     :members:

.. automodule:: pyqalx.core.adapters.item
     :members:

.. automodule:: pyqalx.core.adapters.notification
     :members:

.. automodule:: pyqalx.core.adapters.queue
     :members:

.. automodule:: pyqalx.core.adapters.set
     :members:

.. automodule:: pyqalx.core.adapters.worker
     :members:


Entities
~~~~~~~~
Root entities - these are children of ``dict`` but with special methods.  They
allow for attribute and key lookup

.. _bot_entity:
.. automodule:: pyqalx.core.entities.bot
    :members:

.. _qalx_entity:
.. automodule:: pyqalx.core.entities.entity
    :members:

.. _group_entity:
.. automodule:: pyqalx.core.entities.group
    :members:

.. _item_entity:
.. automodule:: pyqalx.core.entities.item
    :members:

.. _queue_entity:
.. automodule:: pyqalx.core.entities.queue
    :members:

.. _set_entity:
.. automodule:: pyqalx.core.entities.set
    :members:

.. _worker_entity:
.. automodule:: pyqalx.core.entities.worker
    :members:

.. _blueprint_entity:
.. automodule:: pyqalx.core.entities.blueprint
    :members:

Factories
~~~~~~~~~

Factories provide an automated way to create multiple bots - both locally and
on remote servers

.. _factory:
.. automodule:: pyqalx.factories.factory
    :members:

Transport
~~~~~~~~~

The transport layer of ``pyqalx``.  Links adapters to the REST API

.. _transport_api:
.. automodule:: pyqalx.transport.api
    :members:

.. _transport_core:
.. automodule:: pyqalx.transport.core
    :members:

Config
~~~~~~

.. automodule:: pyqalx.config.configs
    :members:

Logging
~~~~~~~

.. automodule:: pyqalx.core.log
    :members:


Signals
~~~~~~~

.. automodule:: pyqalx.core.signals
    :members:

Errors
~~~~~~

.. automodule:: pyqalx.core.errors
    :members:



"""



"""
Custom types module. Defines types that extend the ones available from click
to use for validation of the config options
"""

import logging

from click.types import BoolParamType, Choice, Path, IntRange


class ExtendedBoolType(BoolParamType):
    def convert(self, value, *args):
        """
        Uses the conversion according to the click type, but additionally
        accepts `on/off` options for True and False respectively
        """
        if value == "on":
            return True
        if value == "off":
            return False
        return super().convert(value, *args)


BOOL = ExtendedBoolType()
LOGGING_OPTIONS = Choice(choices=(logging._levelToName.values()))
DIR_PATH_NON_EXISTING = Path(exists=False, dir_okay=True)
FILE_PATH_NON_EXISTING = Path(exists=False, file_okay=True)
FILE_PATH_EXISTING = Path(exists=True, file_okay=True)
POSITIVE_INT = IntRange(min=1)



from treelib import Tree

from pyqalx.core.errors import QalxError


class QalxTree(Tree):
    """
    Given a tree structure as a dictionary will build a treelib.Tree
    instance of that structure.  Implements a few helper methods in order
    to find specific nodes quickly
    """

    def __init__(self, tree_definition, *args, **kwargs):
        super(QalxTree, self).__init__(*args, **kwargs)
        assert isinstance(tree_definition, dict)

        if len(tree_definition) > 1:
            raise QalxError(
                f"Must have only one root node."
                f'  Got {", ".join(tree_definition.keys())}'
            )
        root, successors = tuple(tree_definition.items())[0]
        self.create_node(root, root)
        # Create all the successor nodes recursively
        self.create_successors(root, successors)

    def create_successors(self, root, successor):
        """
        For a given root will create the successor nodes recursively
        :param root: The root node
        :param successor: The successor node
        """
        if isinstance(successor, str):
            # Leaf Node
            self.create_node(successor, successor, parent=root)
        elif isinstance(successor, dict):
            # Dict with successor nodes
            for _successor_name, _successors in successor.items():
                self.create_node(_successor_name, _successor_name, parent=root)
                self.create_successors(_successor_name, _successors)
        elif isinstance(successor, list):
            # List of sibling nodes
            for _ in successor:
                self.create_successors(root, _)

    def get_successors(self, node_name):
        """
        Given a name will provide the immediate successors to this node
        :param node_name: The name of the node that you want to find the
        successors for
        :return:The immediate successors to the found node
        """
        node = self.get_node(node_name)
        if node is not None:
            return node.successors(self.identifier)



from jsonschema import validate, validators

from pyqalx.core.entities import QalxEntity, Item, Set
from pyqalx.core.errors import QalxInvalidBlueprintError


class Blueprint(QalxEntity):
    """
    Provides an interface for creating blueprints.  Blueprints allow a user to
    create items or sets in a consistent fashion
    """

    entity_type = "blueprint"

    def _update_property(self, property_key, key, schema_value, required=False):
        self["properties"][property_key]["properties"][key] = self.__class__(
            schema_value
        )
        if (required is True) and (
            key not in self["properties"][property_key]["required"]
        ):
            self["properties"][property_key]["required"].append(key)
        self.check_schema()
        return self

    @classmethod
    def _base_blueprint(cls):
        return cls(
            {
                "type": "object",
                "properties": {
                    "meta": {"type": "object", "properties": {}, "required": []}
                },
            }
        )

    @classmethod
    def item(cls):
        """
        A blueprint that gives the default blueprint structure for an
        :class:`~pyqalx.core.entities.item.Item` entity
        """
        blueprint = cls._base_blueprint()
        blueprint.entity_type = Item.entity_type
        blueprint.properties.data = cls(
            {"type": "object", "properties": {}, "required": []}
        )
        blueprint.required = ["data"]
        return blueprint

    @classmethod
    def set(cls):
        """
        A blueprint that gives the default blueprint structure for an
        :class:`~pyqalx.core.entities.set.Set` entity
        """
        blueprint = cls._base_blueprint()
        blueprint.entity_type = Set.entity_type
        blueprint.properties["items"] = cls(
            {"type": "object", "properties": {}, "required": []}
        )
        blueprint.required = ["items"]
        return blueprint

    def check_schema(self, schema=None):
        """
        Checks that the given schema is structure correctly

        :param schema: The schema to validate.  Uses `self` if None
        :type schema: None or ~pyqalx.core.entities.blueprint.Blueprint

        :raises: jsonschema.SchemaError
        """
        if schema is None:
            schema = self
        validators._LATEST_VERSION.check_schema(schema=schema)

    def validate(self, entity):
        """
        Validate an entity against the blueprint schema locally.

        :param entity: A full schema for an entity that can have a Blueprint
        """
        validate(instance=entity, schema=self["schema"])

    def add_schema(self, schema_value):
        """
        Update the entire Blueprint schema at once.  Will
        completely overwrite all `properties` that are already on this Blueprint
        instance

        :param schema_value: A full schema for an entity that can have a
                             Blueprint
        :return: An updated instance of the blueprint
        """
        # addict will update keys when merging - not replace.  But
        # we want to replace the properties with the new schema
        # https://github.com/mewwts/addict#update
        self.properties = self.__class__(schema_value)
        self.check_schema()
        return self

    def add_data(self, key, schema_value, required=False):
        """
        Adds data to this item blueprint instance.

        :param key: The key in the data you want to add
        :type key: str
        :param schema_value: The schema value you want to add
        :type schema_value: str
        :param required: Is this key required?
        :type required: bool
        :return: An updated instance of the item blueprint
        """
        return self._update_property(
            property_key="data",
            key=key,
            schema_value=schema_value,
            required=required,
        )

    def add_meta(self, key, schema_value, required=False):
        """
        Adds meta to this item blueprint instance.

        :param key: The key in the meta you want to add
        :type key: str
        :param schema_value: The schema value you want to add
        :type schema_value: str
        :param required: Is this key required?
        :type required: bool
        :return: An updated instance of the item blueprint
        """
        return self._update_property(
            property_key="meta",
            key=key,
            schema_value=schema_value,
            required=required,
        )

    def add_item(self, key, item_blueprint=None, required=False):
        """
        Adds an item to this set blueprint instance.  If `item_blueprint` is
        provided then the set schema will always require that specific
        item blueprint for `key`.  If `item_blueprint` is not provided then
        the set blueprint will accept any data for `key` when being used

        :param key: The `key` of the item you want the blueprint to validate
        :type key: str
        :param item_blueprint: If provided will save the `name` from the
                               blueprint against the `key`.
                               This means that whenever the set is
                               created you must use the same blueprint.
        :type item_blueprint: str
        :param required: Is this key required?
        :type required: bool
        :return: An updated instance of the set blueprint
        """
        if item_blueprint is None:
            # the data stored against `key` can be anything when the set is
            # created
            schema_value = {"type": "object"}
        else:
            # the data stored against `key` must match the item blueprint
            if not isinstance(item_blueprint, self.__class__):
                raise QalxInvalidBlueprintError(
                    f"`{item_blueprint}` is not an"
                    f" instance of `{self.__class__}`"
                )
            # The "name" is required in order to pack the Blueprint
            # in QalxBlueprint for saving.
            item_blueprint.schema.name = item_blueprint["name"]
            schema_value = item_blueprint["schema"]
        return self._update_property(
            property_key="items",
            key=key,
            schema_value=schema_value,
            required=required,
        )



from .entity import QalxEntity, QalxListEntity  # noqa: F401
from .bot import Bot  # noqa: F401
from .group import Group  # noqa: F401
from .item import Item  # noqa: F401
from .set import Set  # noqa: F401
from .queue import Queue  # noqa: F401
from .blueprint import Blueprint  # noqa: F401
from .notification import Notification  # noqa: F401



from collections import namedtuple

from pyqalx.core.adapters.adapter import QalxAdapter
from pyqalx.core.errors import QalxInvalidTagError, QalxAPIResponseError


class Tags(list):
    """
    Tags are a list of Tag() instances.

    These tags are then used on every `add` and `save` request for the
    duration of the session.
    Each individual tag is instantiated as a Tag instance to handle validation
    to ensure they are in the correct format.
    To add new tags to a session do `session.tags.add(name='<name>',
                                                      value='<value>')`
    """

    def __init__(self, *args, **kwargs):
        # rest_api is required to validate tags
        self.rest_api = kwargs.pop("rest_api")

        super(Tags, self).__init__(*args, **kwargs)

        self._make_iterable(self)

    def _make_tag(self, name, value, validate=True):
        """
        Helper method for converting a name & value into a namedtupled and
        then to a dictionary to be added to `Tags`
        :param name: The value for the `name` key on a Tag
        :param value: The value for the `value` key on a Tag
        :return: A dict
        """
        tag = dict(
            Tag(
                name=name.lower().strip(), value=value.lower().strip()
            )._asdict()
        )
        if validate:
            # We don't always validate (i.e. if this was called from
            # `_make_iterable`) as we can validate all tags in a single request
            # on the `_make_iterable` method.
            self._validate(tag)
        return tag

    def _make_iterable(self, iterable):
        """
        Mutates the given iterable to ensure that the keys on it our
        valid for a `Tag`
        """
        for index, tag in enumerate(iterable):
            iterable[index] = self._make_tag(validate=False, **tag)

        self._validate(iterable)

    def _validate(self, tags):
        """
        Makes an API request that validates that the user has
        permission to write to the given tags
        """
        if not isinstance(tags, list):
            tags = [tags]
        if len(tags):
            # Non standard endpoint so cannot use a QalxSession instance
            success, data = self.rest_api.post(
                "tags/validate", json={"tags": tags}
            )
            if not success:
                try:
                    QalxAdapter.raise_api_response_error(data)
                except QalxAPIResponseError as exc:
                    raise QalxInvalidTagError(exc)

    def add(self, name, value):
        """
        A helper method to make it easier for users to add tags.
        :param name: The name of the tag
        :param value: The value of the tag
        """
        super(Tags, self).append(self._make_tag(name=name, value=value))

    def append(self, p_object):
        super(Tags, self).append(self._make_tag(**p_object))

    def insert(self, index, p_object):
        super(Tags, self).insert(index, self._make_tag(**p_object))

    def extend(self, iterable):
        self._make_iterable(iterable)
        super(Tags, self).extend(iterable)

    def __add__(self, iterable):
        self._make_iterable(iterable)
        super(Tags, self).__add__(iterable)

    def __iadd__(self, iterable):
        self._make_iterable(iterable)
        super(Tags, self).__iadd__(iterable)

    def _query(self):
        """
        Returns the tags in a format for querying
        """
        return [
            {
                "tags": {
                    "$elemMatch": {
                        "name": p["name"].lower().strip(),
                        "value": p["value"].lower().strip(),
                    }
                }
            }
            for p in self
        ]


# An individual tag is just a namedtuple with a "name" and "value" key.
# These get converted to `OrderedDict` when being added to `Tags`
Tag = namedtuple("Tag", ["name", "value"])



from .blueprint import QalxBlueprint  # noqa: F401
from .bot import QalxBot  # noqa: F401
from .group import QalxGroup  # noqa: F401
from .item import QalxItem  # noqa: F401
from .queue import QalxQueue  # noqa: F401
from .set import QalxSet  # noqa: F401
from .worker import QalxWorker  # noqa: F401
from .factory import QalxFactory  # noqa: F401
from .notification import QalxNotification  # noqa: F401



from copy import deepcopy

from pyqalx.core.adapters.adapter import (
    QalxNamedEntityAdapter,
    QalxUnpackableAdapter,
)
from pyqalx.core.entities.blueprint import Blueprint
from pyqalx.core.errors import QalxInvalidBlueprintError


class QalxBlueprint(QalxUnpackableAdapter, QalxNamedEntityAdapter):
    """
    Provides an interface for accessing the API for
    :class:`~pyqalx.core.entities.blueprint.Blueprint` entities
    """

    _entity_class = Blueprint
    child_entity_class = Blueprint

    def _validate(self, **kwargs):
        """
        Validates that the blueprint is in the correct format, handling
        potential issues if it is a nested schema
        """
        # The _kids_for_lookup method does everything we need in order to
        # determine the nested kids, so just call that with validate=True
        self._kids_for_lookup(entity=kwargs)
        return super(QalxBlueprint, self)._validate(**kwargs)

    def _check_schema(self, schema):
        self._entity_class({}).check_schema(schema)

    def add(self, name, schema, meta=None, **kwargs):
        """
        Adds a :class:`~pyqalx.core.entities.blueprint.Blueprint` with a
        valid `jsonschema` `schema`.
        If the `schema` is invalid then a `jsonschema.SchemaError` is raised

        :param name: The name of this blueprint
        :type name: str
        :param schema: The schema that you want to set on the Blueprint.
        :type schema: dict
        :param meta: A dictionary of metadata to store
        :type meta: dict

        :return: :class:`~pyqalx.core.entities.blueprint.Blueprint`
        :raises: jsonschema.SchemaError
        """
        self._check_schema(schema)
        entity_type = schema.get("entity_type", None)
        if entity_type is None:
            raise QalxInvalidBlueprintError(
                "schema must specify "
                "`entity_type` top level key which "
                "has a valid `entity_type` as the "
                "value"
            )
        return super(QalxBlueprint, self).add(
            name=name,
            schema=schema,
            entity_type=entity_type,
            meta=meta,
            **kwargs,
        )

    def save(self, entity, **kwargs):
        """
        Saves any updates to the given
        :class:`~pyqalx.core.entities.blueprint.Blueprint`.  Validates that
        the `schema` on the entity is a valid `jsonschema` schema

        :param entity: A valid Blueprint instance
        :type entity: ~pyqalx.core.entities.blueprint.Blueprint
        :param kwargs: Any kwargs you want to save against this `Blueprint`
        :return: :class:`~pyqalx.core.entities.blueprint.Blueprint`
        :raises: jsonschema.SchemaError
        """
        if "schema" in entity.keys():
            self._check_schema(entity["schema"])
        return super(QalxBlueprint, self).save(entity=entity, **kwargs)

    def get(self, guid, child_fields=None, unpack=True, *args, **kwargs):
        # extended to create better autodocs
        return super(QalxBlueprint, self).get(
            guid=guid, child_fields=child_fields, unpack=unpack, *args, **kwargs
        )

    def get_by_name(self, name, **kwargs):
        # extended to create better autodocs
        return super(QalxBlueprint, self).get_by_name(name=name, **kwargs)

    def get_or_create(self, name, meta=None, **kwargs):
        # extended to create better autodocs
        return super(QalxBlueprint, self).get_or_create(
            name=name, meta=meta, **kwargs
        )

    def find(
        self,
        query=None,
        sort=None,
        skip=0,
        limit=25,
        many=True,
        child_fields=None,
        include_session_tags=True,
        *args,
        **kwargs,
    ):
        """
        Return multiple packed entities from the API

        :param query: The optional Mongo query to find entities
        :type query: dict
        :param sort: The keys to sort by
        :type sort: list
        :param skip: The number of results to skip (offset) by
        :type skip: int
        :param limit: How many results should the response be limited to
        :type limit: int
        :param many: Should many entities be returned or just a single one
        :type many: bool
        :param include_session_tags:    Should the tags for the session be
                                        included in the query. Default is True
        :type include_session_tags: bool
        :param child_fields: A list of fields that should be returned from
                             child entities
        :type child_fields: list
        :return: A list of entities
        """
        # extended to create better autodocs
        return super(QalxBlueprint, self).find(
            query=query,
            sort=sort,
            skip=skip,
            limit=limit,
            many=many,
            include_session_tags=include_session_tags,
            child_fields=child_fields,
            *args,
            **kwargs,
        )

    def find_one(self, query, unpack=True, **kwargs):
        """
        Method for returning a unique entity. Will return the entity that
        matches the query

        :param query: The mongoDB query
        :type query: dict
        :param unpack: Should any child entities be automatically unpacked?
        :type unpack: bool
        :return: an instance of `self.entity_class`
        :raises QalxMultipleEntityReturned:
        :raises QalxEntityNotFound:
        """
        # extended to create better autodocs
        return super(QalxBlueprint, self).find_one(
            query, unpack=unpack, **kwargs
        )

    def reload(self, entity, unpack=True, *args, **kwargs):
        """
        Reloads the current entity from the API

        :param entity: An instance of Blueprint
        :type entity: ~pyqalx.core.entities.blueprint.Blueprint
        :param unpack: Should any child entities be automatically unpacked?
        :type unpack: bool
        :return: A refreshed instance of `self._entity_class`
        """
        # extended to create better autodocs
        return super(QalxBlueprint, self).reload(
            entity=entity, unpack=unpack, *args, **kwargs
        )

    def aggregate(self, aggregate, **kwargs):
        # extended to create better autodocs
        return super(QalxBlueprint, self).aggregate(
            aggregate=aggregate, **kwargs
        )

    @staticmethod
    def _kids_properties(entity):
        """
        A Blueprint may not be nested (i.e. if it's an Item Blueprint).
        Therefore, check get the kids properties by checking if the "schema"
        key is there (may not be if "schema" wasn't returned from the API,
        and check that `kids` is there (may not be if not nested blueprint)
        :param entity: An instance of Blueprint
        :return: Boolean determining if this Blueprint is nested
        """
        schema = entity.get("schema", {}).get("properties", {})
        # currently the only nested blueprint can be Sets. Which
        # will have the kids nested using the `items` key
        kids = schema.get("items", {}).get("properties")
        return kids

    def _kids_for_lookup(self, entity):
        """
        The blueprint kids will be nested deeply in the schema as enums.
        We should iterate through them (checking to make sure each nested
        key only has a single enum) and return them
        :param entity: The blueprint entity
        :type entity: instance of Blueprint
        :return:The child blueprint names as list
        """
        blueprint_names = []
        if self._kids_properties(entity):
            # Build the blueprint names to lookup - handling either no
            # blueprint name or too many blueprint names
            for key, value in self._kids_properties(entity).items():
                child_blueprint_name = value.get("enum")
                if child_blueprint_name:
                    if len(child_blueprint_name) > 1:
                        # Can only happen if a user builds a schema themselves
                        raise QalxInvalidBlueprintError(
                            f"Multiple `enum` values "
                            f"found for `{key}`. "
                            f"Ensure only one "
                            f"is specified`"
                        )
                    blueprint_names.append(child_blueprint_name[0])
        return blueprint_names

    def _child_list_request(
        self, entity, kids, child_adapter, query_key="name", **kwargs
    ):
        """
        We lookup a blueprints kids using the "name" query key rather than
        the guid as default as the kids are stored as names on the API
        :param entity: Instance of Blueprint
        :type entity: ~entities.blueprint.Blueprint
        :param kids: A list of blueprint names
        :type kids: list
        :param child_adapter: The child adapter
        :type child_adapter: Instance of QalxBlueprint
        :param query_key: The query key
        :type query_key: str
        :return: The unpacked child blueprint entities
        """
        return super(QalxBlueprint, self)._child_list_request(
            entity=entity,
            kids=kids,
            child_adapter=child_adapter,
            query_key=query_key,
            **kwargs,
        )

    def _pack_kids_for_request(self, kwargs):
        """
        If the Blueprint is a nested blueprint, will pack the kids to be
        enums rather than the fully inflated kids schema that would have come
        from the Blueprint entity
        """
        if self._kids_properties(kwargs):
            for key, value in self._kids_properties(kwargs).items():
                if "name" in value:
                    kwargs["schema"]["properties"]["items"]["properties"][
                        key
                    ] = {"type": "string", "enum": [value["name"]]}
                else:
                    if value != {"type": "object"}:
                        # A name wasn't specified and the nested entity
                        # has been defined.  This can only happen if a user
                        # builds the schema manually.  We require the `name`
                        # field in order to build the `enum` list to enable
                        # packing/unpacking to work.
                        raise QalxInvalidBlueprintError(
                            f"{key} key must specify `name` key which should "
                            f"be the name of the specific child blueprint. "
                            f"This is required for packing/unpacking to work."
                        )
        return kwargs

    def _attempt_unpack(self, entity, **kwargs):
        """
        Have to override _attempt_unpack because the `self.kids` key is not
        actually where we need to get the children from.
        """
        if self._kids_properties(entity):
            unpacked_entities = self._get_child_entities(entity, **kwargs)
            self._unpacked_entities_to_valid_children(
                entity=entity, unpacked_entities=unpacked_entities
            )
        return entity

    def cache_kids(self, kwargs):
        """
        Override `cache_kids` because the `self.kids` key is not where
        we need to get the children from. Using deepcopy because schema is a
        mutable object(dict)
        """
        return deepcopy(kwargs["schema"])

    def save_kids_on_entity(self, entity, kids):
        """
        Override `save_kids_on_entity` because the `self.kids` key is
        not where the children should be saved.
        """
        entity["schema"] = kids

    def _unpacked_entities_to_valid_children(self, entity, unpacked_entities):
        """
        Iterates the enum values of the child blueprints and replaces them
        with the full item schema
        """
        kids_properties = self._kids_properties(entity)
        for key, value in kids_properties.items():
            for child_entity in unpacked_entities:
                if (
                    value.get("enum")
                    and value.get("enum")[0] == child_entity.name
                ):
                    kids_properties[key] = child_entity.schema
        return kids_properties



"""
Core API
~~~~~~~~

.. automodule:: pyqalx.core
    :members:

Bot API
~~~~~~~

.. automodule:: pyqalx.bot
    :members:

"""
from .core.adapters import (  # noqa: F401
    QalxItem,
    QalxSet,
    QalxGroup,
    QalxQueue,
    QalxBot,
)
from .core.session import QalxSession  # noqa: F401
from .bot import Bot  # noqa: F401
from .core.entities import Blueprint, Item, Set, Group, Queue  # noqa: F401



from pyqalx.core.errors import QalxError


class QalxSignal(object):
    """
    A set of helper methods to aid in managing signals on Workers
    """

    BASE_EXIT_CODE = 72590
    TERMINATE_WARM_CODE = 72591
    TERMINATE_COLD_CODE = 72592
    CONFIG_TERMINATE_CODE = 72593
    RESUME_CODE = 72594
    STOP_CODE = 72595
    COLD_EXIT_KEY = "cold"
    WARM_EXIT_KEY = "warm"
    REQUEUE_JOB_KEY = "requeue_job"
    STOP_KEY = "stop"
    TERMINATE_KEY = "terminate"

    def __init__(self, entity):
        self.signal = entity["signal"]

    @classmethod
    def _build_signal(cls, key, value):
        """
        Builds the signal structure.  A signal should be a dict in the format
        {<key>: <bool>}
        """
        return {key: value}

    @classmethod
    def _terminate_signal(cls):
        return cls._build_signal(cls.TERMINATE_KEY, True)

    @classmethod
    def _terminate_warm_signal(cls):
        return cls._build_signal(cls.WARM_EXIT_KEY, True)

    @classmethod
    def _stop_signal(cls, stop):
        return cls._build_signal(cls.STOP_KEY, stop)

    @classmethod
    def _build_update(cls, update, code):
        """
        Builds an update to publish to the worker signal

        :param update: The update string
        :param code: The exit code
        """
        return {"update": update, "code": code}

    @classmethod
    def _warm_exit_update(cls):
        return cls._build_update(
            f"{cls.WARM_EXIT_KEY} exit signal found", cls.TERMINATE_WARM_CODE
        )

    @classmethod
    def _cold_exit_update(cls):
        return cls._build_update(
            f"{cls.COLD_EXIT_KEY} exit signal found", cls.TERMINATE_COLD_CODE
        )

    @classmethod
    def _config_exit_update(cls):
        return cls._build_update(
            "KILL_AFTER reached", cls.CONFIG_TERMINATE_CODE
        )

    @classmethod
    def _stop_or_resume_update(cls, stop):
        return cls._build_update(
            f"{'stop' if stop else 'resume'} signal found",
            cls.STOP_CODE if stop else cls.RESUME_CODE,
        )

    @property
    def terminate(self):
        """
        Checks whether the signal was a terminate signal
        """
        return self.signal.get(self.TERMINATE_KEY)

    @property
    def terminate_cold(self):
        """
        Checks whether the signal was a cold terminate signal
        """
        return bool(self.terminate and self.signal.get(self.COLD_EXIT_KEY))

    @property
    def terminate_warm(self):
        """
        Checks whether the signal was a warm terminate signal
        """
        return bool(self.terminate and self.signal.get(self.WARM_EXIT_KEY))

    @property
    def requeue_job(self):
        """
        Checks whether the signal wants the job to be requeued
        """
        return bool(self.signal.get(self.REQUEUE_JOB_KEY))

    @property
    def stop(self):
        """
        Checks whether the signal wants to stop the job
        """
        return bool(self.signal.get(self.STOP_KEY))


class QalxWorkerSignal(QalxSignal):
    """
    Handles signals for individual instances of :class:`~pyqalx.bot.Worker`
    """

    @classmethod
    def _terminate_cold_signal(cls, requeue_job):
        cold = cls._build_signal(cls.COLD_EXIT_KEY, True)
        requeue = cls._build_signal(cls.REQUEUE_JOB_KEY, bool(requeue_job))
        return {**requeue, **cold}

    @classmethod
    def _terminate_signal(cls, warm=True, requeue_job=None, **kwargs):
        """
        Workers can either warm or cold terminate.

        :param warm: Whether this should be a warm or cold termination
        :param requeue_job: Whether we should requeue the job (cold terminate
                            only)
        :param kwargs:
        :return:
        """
        signal = super(QalxWorkerSignal, cls)._terminate_signal()
        if warm and requeue_job:
            raise QalxError(
                "warm termination cannot be called with requeue_job, you "
                "need to add the job to the queue manually."
            )
        if warm:
            extra = cls._terminate_warm_signal()
        else:
            extra = cls._terminate_cold_signal(requeue_job=requeue_job)
        return {**signal, **extra}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyqalx',
 'pyqalx.bot',
 'pyqalx.config',
 'pyqalx.config.defaults',
 'pyqalx.core',
 'pyqalx.core.adapters',
 'pyqalx.core.entities',
 'pyqalx.factories',
 'pyqalx.factories.sectors',
 'pyqalx.factories.sectors.providers',
 'pyqalx.factories.sectors.providers.aws',
 'pyqalx.factories.sectors.providers.local',
 'pyqalx.transport',
 'pyqalx.vendor',
 'qalxcli',
 'qalxcli.cli']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2<3.1',
 'addict>=2.2,<3.0',
 'appdirs>=1.4,<2.0',
 'boto3>=1.12.31,<2.0.0',
 'bson-extra>=0.0.4,<0.0.5',
 'click>=8.0,<9.0',
 'colorama>=0.4.3,<0.5.0',
 'cryptography>=35.0.0,<36.0.0',
 'deprecation>=2.0,<3.0',
 'dill>=0.3.3',
 'gitpython>=3.1,<4.0',
 'jsonschema>=3.0,<4.0',
 'pyyaml>=5.3,<6.0',
 'requests>=2.21,<3.0',
 'tabulate>=0.8.10,<0.9.0',
 'treelib>=1.6,<2.0',
 'troposphere>=3.1.1,<4.0.0']

extras_require = \
{'flaky': ['flaky>=3.5,<4.0'],
 'pip-licenses': ['pip-licenses>=3.5.3,<4.0.0'],
 'sphinx': ['sphinx>=2.0,<3.0']}

entry_points = \
{'console_scripts': ['qalx = qalxcli.cli.qalx:qalx']}

setup_kwargs = {
    'name': 'pyqalx',
    'version': '0.19.0',
    'description': 'High-level interfaces to the qalx API',
    'long_description': 'pyqalx\n======\n.. image:: https://img.shields.io/bitbucket/pipelines/agiletekengineering/pyqalx   :alt: Bitbucket Pipelines\n\n\n.. image:: https://api.codacy.com/project/badge/Grade/fdfb81fd286a474098b624b426d63c41\n    :target: https://www.codacy.com?utm_source=bitbucket.org&amp;utm_medium=referral&amp;utm_content=agiletekengineering/pyqalx&amp;utm_campaign=Badge_Grade\n\n\n.. image:: https://api.codacy.com/project/badge/Coverage/fdfb81fd286a474098b624b426d63c41\n    :target: https://www.codacy.com?utm_source=bitbucket.org&amp;utm_medium=referral&amp;utm_content=agiletekengineering/pyqalx&amp;utm_campaign=Badge_Coverage\n\n\nInterfaces to qalx. For more details, see `project documentation, <http://docs.qalx.net>`_.\n\n.. admonition:: development status\n\n   ``pyqalx`` is currently under active development. It is pre-version 1.0 beta software and so each minor version\n   can introduce breaking changes.\n\n**qalx** (an abbreviation of "queued calculations" and pronounced "kal-x") is a flexible data management platform for engineering projects. Users store data and files in qalx and it provides tools for passing that data between various systems for processing.\n\nThere will eventually be four ways to interact with the platform:\n\n-  A Python interface (pyqalx)\n-  REST API (api.qalx.net)\n-  Web console (console.qalx.net) - coming soon\n-  A command line interface (qalx-cli) - coming soon\n\nMost users are expected to use the web console and either the python or\ncommand line interfaces. The REST API is intended to be used if you\nprefer to access the platform with a language other than Python or want\nto create a custom interface.\n\n.. _installing:\n\nInstalling\n==========\n\n**qalx** is written in `Python <https://python.org>`_ and can be\ninstalled via the Python Package Index (PyPi) with:\n\n.. code:: bash\n\n   pip install pyqalx\n\nIf installation has completed properly you should be able to import\n``pyqalx`` in a python console:\n\n>>> import pyqalx\n\n.. warning::\n\n      pyqalx requires **Python versions above 3.6**.\n\n\nConfiguration and Authentication\n--------------------------------\n\nEverything you do with **qalx** requires you to be authenticated. That\nis, the platform requires you to identify yourself and will record all\nyour actions as being performed by you.\n\nThe way that **qalx** knows who you are is by reading a ``TOKEN``\nwhich must be sent with every request.\n\n.. warning::\n   During this beta phase, you have to request a ``TOKEN`` by registering your interest at `qalx.net <https://qalx.net>`_\n\nThe easiest way to make sure that your token is sent with every request is to make sure you have a valid ``.qalx`` file\nsaved in your HOME directory.\n\n.. admonition::  where is ``HOME``?\n\n   The ``HOME`` directory can usually be found by putting %USERPROFILE%\n   in the address bar in Windows Explorer or it is simply ``~`` on unix\n   systems.\n\n\nqalx CLI configure\n------------------\nYou can either add the ``TOKEN`` manually to the config file or you can use the **qalx** cli to do this for you automatically\n\n.. code:: bash\n\n   $ qalx configure\n\nThe above command will ask you all the necessary questions to create your user and bot config files.\n\nYou can also create other profiles using the ``qalx configure`` command. You can even create profiles just for bots\nor just for users - or provide extra arguments to write to the config\n\n.. code:: bash\n\n   # Will write a user profile called `dev` to the config file,\n   # will also write a `default` bot config profile\n   $ qalx --user-profile dev configure\n\n   # Will write a bot profile called `dev` to the config file,\n   # will also write a `default` user config profile\n   $ qalx --bot-profile dev configure\n\n   # Will not write the bot profile\n   $ qalx configure --no-bot\n\n   # Will not write the user profile\n   $ qalx configure --no-user\n\n   # Will write a default user and bot profile and will also write two\n   # extra keys to each config - `CUSTOMKEY=customvalue`\n   # and `CUSTOMKEY2=customvalue2`\n   $ qalx configure customkey=customvalue customkey2=customvalue2\n\nManual configuration\n--------------------\n\nIf you don\'t want to use the ``qalx cli`` to configure your profile (or if you want to make changes to it in the future)\nthen you can configure the profile yourself by adding the ``TOKEN`` to your config file under\nthe default profile.\n\n.. code:: ini\n\n   [default]\n   TOKEN = 632gd7yb9squd0q8sdhq0s8diqsd0nqsdq9sdk\n\nAny other configuration settings can be stored in the same file see `<configuration>`_ for more information.\n\nQuickstart\n----------\n\nThe best place to get started: `<https://docs.qalx.net/quickstart>`_\n\n\n\n',
    'author': 'AgileTek Engineering Limited',
    'author_email': 'london@agiletek.co.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://qalx.net',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from pyqalx.core.entities.set import Set
from pyqalx.core.entities.entity import QalxQueueableEntity


class Group(QalxQueueableEntity):
    """A group is a collection of `sets
    <https://en.wikipedia.org/wiki/Turtles_all_the_way_down>`_.
    These are useful for sending a lot of sets to a queue and being able to
    track when they have all been processed.


    >>> from pyqalx import QalxSession
    >>> qalx = QalxSession()
    >>> steel = qalx.item.add(source="path/to/316_datasheet.pdf",
    ...                       data={"rho":8000, "E":193e9},
    ...                       meta={"library":"materials", "family":"steel",
    ...                             "grade":"316"})
    >>> steel_squares = {}
    >>> for size in range(2, 500):
    ...     dims = qalx.item.add(data={"height":size, "width":size},
    ...                          meta={"shape":"square"})
    ...     steel_square_set = qalx.set.add(items={"shape":dims, "material":steel},
    ...                                     meta={"profile":"square_steel"})
    ...     steel_squares[size] = steel_square_set
    >>> all_squares = qalx.group.add(sets=steel_squares)
    """

    entity_type = "group"
    child_entity_class = Set



from pyqalx.core.adapters.adapter import QalxAdapter
from pyqalx.core.entities import Notification


class QalxNotification(QalxAdapter):
    """
    Provides an interface for accessing the API for
    :class:`~pyqalx.core.entities.notification.Notification` entities
    """

    _entity_class = Notification

    def add(self, subject, message, to, cc=None, bcc=None, **kwargs):
        """
        Posts a notification request to the API, to send an email with the
        provided subject, message and recipient list. Optionally, lists for CC
        and BCC recipients can be provided

        :param subject: The subject of the message
        :type subject: str
        :param message: The message to send
        :type message: str
        :param to: A list with the addresses of the recipients of the message
        :type to: list
        :param cc:  An optional list with the addresses of th recipients of a
                    carbon copy(CC) of the message
        :type cc: list
        :param bcc:  An optional list with the addresses of th recipients of a
                    blind carbon copy(BCC) of the message
        :type bcc: list
        """
        if cc is None:
            cc = []
        if bcc is None:
            bcc = []
        return super(QalxNotification, self).add(
            subject=subject, message=message, to=to, cc=cc, bcc=bcc, **kwargs
        )



import json
import threading
from datetime import datetime, timezone
import uuid

import boto3
from botocore.exceptions import ClientError

from pyqalx.core.entities import QalxEntity, Group, Item, Set
from pyqalx.core.errors import QalxQueueError


class QueueResponse:
    """a response from a remote queue"""

    def __init__(self, response):
        """make new response

        :param response: response from queue
        """
        self._raw_response = response


class QueueMessage(threading.Thread):
    """a message from the queue.  subclasses `threading.Thread` so that a
    heartbeat can be started as soon as the message is received from the Queue
    """

    def __init__(self, message, worker, visibility, *args, **kwargs):
        """
        :param message: the full raw message from the queue
        :type message: dict
        :param worker: An instance of :class:~pyqalx.bot.Worker
        :type worker: ~pyqalx.bot.Worker
        :param visibility: How long should the heartbeat "beat" for (in seconds)?
        :type visibility: int
        """
        super(QueueMessage, self).__init__(*args, **kwargs)
        self._raw_message = message
        self.body = json.loads(message["Body"])
        self.receipt_handle = message["ReceiptHandle"]

        # Set up the heartbeat event and start the thread.
        # This will "beat" every 50% of visibility seconds and increase
        # the visibility timeout of the message on the queue.
        self._heartbeats = 1
        self._heartbeat = threading.Event()
        self._worker = worker
        self._log = self._worker.log
        self._queue = self._worker.queue
        self._visibility = visibility
        # Start the threaded heartbeat as soon as we can to prevent
        # message timeouts expiring if it takes a while to start processing them
        self.start()

    def _stop_heartbeat(self):
        """
        Stops any heartbeat that might be mid beat.  At this point the message
        should be fully processed or the max visibility has been reached.
        """
        self._heartbeat.set()

    def _increase_visibility_timeout(self, visibility_timeout):
        """
        Helper for increasing the visibility timeout of this message.  If the
        timeout is set to > MAX_VISIBILITY_TIMEOUT an error will be logged and
        the timeout will be set to MAX_VISIBILITY_TIMEOUT
        :param visibility_timeout: The visibility timeout the message should
        have. This will set the timeout to be a total number of seconds from
        when the message first went in flight
        :type visibility_timeout: int
        """
        MAX_VISIBILITY_TIMEOUT = 43200  # 12 hours
        if visibility_timeout > MAX_VISIBILITY_TIMEOUT:
            # SQS doesn't recalculate the max timeout if it were to be given
            # a value greater than the maximum.
            visibility_timeout = MAX_VISIBILITY_TIMEOUT
            docs = "https://docs.qalx.net/queues#in-flight-maximum"
            msg = (
                f"Max visibility timeout of {MAX_VISIBILITY_TIMEOUT} reached."
                f"The job will go back onto the queue when "
                f"this timeout expires. See {docs} for more information. "
            )
            self._log.warning(msg)
            # Stop the heartbeat, it cannot be increased any further than
            # MAX_VISIBILITY_TIMEOUT.
            self._stop_heartbeat()

        try:
            self._queue.broker_client.change_message_visibility(
                QueueUrl=self._queue["info"]["queue_url"],
                ReceiptHandle=self.receipt_handle,
                VisibilityTimeout=visibility_timeout,
            )
        except ClientError as exc:
            # ClientError: There was an issue changing the visibility.
            #              This could be a networking issue or the message no
            #              longer exists on the queue
            self._log.debug(
                f"Received ClientError when extending heartbeat."
                f"  Error was: `{exc}`"
            )
            # Not ideal but the only way to determine from the response
            # if the message is deleted.
            message_deleted = (
                "Message does not exist" in exc.response["Error"]["Message"]
            )

            if message_deleted:
                # The message was deleted from the queue.  The heartbeat should
                # have been stopped but if the user deleted the message in a
                # non standard way it may not have been.  Therefore, make
                # sure the heartbeat is stopped to prevent spamming the logs.
                self._stop_heartbeat()
            return False
        return True

    def run(self):
        """
        A QalxJob might take a long time to process.  Therefore, we
        use a heartbeat on the message to keep it in flight for as long
        as possible.  This runs in a Thread
        on :class:`~pyqalx.core.entities.queue.QueueMessage` instantiation
        to ensure that the message timeout doesn't expire before processing
        even begins (i.e. if a batch of messages was returned)

        As per the SQS docs `here
        <https://docs.aws.amazon.com/AWSSimpleQueueService
        /latest/SQSDeveloperGuide/sqs-visibility-timeout.html>`_

        This will specify the initial visibility timeout (for example, 2 minutes)
        and then — as long as your consumer still works on
        the message — keep extending the visibility timeout by 2
        minutes every minute.
        This uses `visibility` as the initial visibility timeout so every
        `visibility / 2` seconds will increase the timeout by
        `visibility * number of hearbeats` from the initial visibility time.
        """
        while self._heartbeat.is_set() is False:
            # We only do the heartbeat if it hasn't been set (meaning the
            # heartbeat was stopped due to message deletion)

            # Ensure it's an int - it might not be if it's come from the config
            if self._heartbeats != 1:
                # No need to change visibility on the first heartbeat as the
                # message will have the default timeout still.
                visibility_timeout = self._visibility * self._heartbeats

                success = self._increase_visibility_timeout(visibility_timeout)
                if not success:
                    # It wasn't successful, so just return without waiting the
                    # usual amount of time so that the thread can try again.
                    # Wait for a single second to help prevent multiple
                    # requests being fired to AWS
                    self._heartbeat.wait(1)
                    continue

            # Get roughly 50 percent of the default visibility time.
            # Sleep until that time expires and then increase the
            # message visibility the next time this method is called
            self._heartbeat.wait(self._visibility / 2)
            self._worker._check_queue_credentials()

            # We keep track of the number of heartbeats this message has had
            # and increase the timeout by `self._visibility * self._heartbeats`
            # for each heartbeat (after the first one)
            self._heartbeats += 1


class Queue(QalxEntity):
    """QalxEntity with entity_type Queue"""

    entity_type = "queue"

    @staticmethod
    def _stringify(message: dict) -> str:
        return json.dumps(message)

    def __init__(self, *args, **kwargs):
        super(Queue, self).__init__(*args, **kwargs)
        self._broker_client = None

    def _reserved_keys(self):
        reserved_keys = super(Queue, self)._reserved_keys()
        reserved_keys += ["_broker_client"]
        return reserved_keys

    @property
    def broker_client(self):
        """
        An authenticated client to communicate with the remote message broker.

        Windows can't pickle the client if it gets set on __init__ therefore,
        we set it to an internal attribute and cache it for the duration of
        the Queue to avoid getting the client multiple times
        """
        if (
            self._broker_client is None
            and hasattr(self, "info")
            and ("credentials" in self.info)
        ):

            _broker_client = boto3.client(
                "sqs",
                region_name=self._region,
                aws_access_key_id=self.info.credentials.ACCESS_KEY_ID,
                aws_secret_access_key=self.info.credentials.SECRET_ACCESS_KEY,
                aws_session_token=self.info.credentials.SESSION_TOKEN,
            )
            self._broker_client = _broker_client
        return self._broker_client

    def _credentials_expiring_soon(self):
        """
        Returns True if the credentials are expiring soon.  The expiration
        datetime will be a UTC datetime object
        """
        expiration = self.info.credentials.EXPIRATION
        # Refresh when 1 hour or less left on the expiration token
        REFRESH_LIMIT_HOURS = 3600
        return (
            expiration - datetime.now(timezone.utc)
        ).seconds <= REFRESH_LIMIT_HOURS

    def get_messages(self, max_num_msg, visibility, waittime, worker):
        """
        get messages from the queue

        :param max_num_msg: maximum number to retrieve
        :type max_num_msg: int
        :param visibility: time in seconds until the message becomes visible
                           again on the queue
        :type visibility: int
        :param waittime: time to wait for a message
        :type waittime: int
        :param worker: The :class:`~pyqalx.bot.Worker` instance that called this method
        :type worker: ~pyqalx.bot.Worker
        :return: List of :class:`~pyqalx.core.entities.queue.QueueMessage`
        """

        responses = self.broker_client.receive_message(
            QueueUrl=self.info.queue_url,
            MaxNumberOfMessages=max_num_msg,
            VisibilityTimeout=visibility,
            WaitTimeSeconds=waittime,
        )
        if responses.get("Messages"):
            return [
                QueueMessage(
                    message=response, worker=worker, visibility=visibility
                )
                for response in responses["Messages"]
            ]
        return []

    def delete_message(self, message):
        """remove message from the queue

        :param message: message to delete
        :type message: :class:`~pyqalx.core.entities.queue.QueueMessage`
        """
        if message is not None:
            # message could be None if the job it's been
            # on has been told to wait - therefore, don't attempt delete
            self.broker_client.delete_message(
                QueueUrl=self.info.queue_url,
                ReceiptHandle=message.receipt_handle,
            )
            message._stop_heartbeat()

    def _send_message(self, payload, entity_type, **message_kwargs):
        """
        Batch submits the given payload to the instance of Queue.  Figures out
        whether the payload is a subclass of QalxEntity, a string or a list of
        a combination.  Typically this will be invoked via the entity class itself:

        >>> EntityClass.add_to_queue(entity, queue)
        >>> queue._send_message(<entity_instance>)
        >>> queue._send_message(<entity_guid>)
        >>> queue._send_message([<entity_instance>,
        ...                      <entity_instance>,
        ...                      <entity_instance>])
        >>> queue._send_message([<entity_guid>, <entity_guid>, <entity_guid>])

        :param payload: A subclassed instance of QalxEntity, a guid string or a list
                        of QalxEntities/guids
        :type payload:QalxEntity or str
        :param entity_type:The entity type
        :type entity_type: str
        :param message_kwargs:Additional arguments that should be added to the
                              message body
        :return:An instance of QueueResponse
        """

        responses = []
        valid_types = (QalxEntity, uuid.UUID, list)
        if not isinstance(payload, valid_types):
            # Only certain types are supported
            raise QalxQueueError(
                f"Expected one of {', '.join([str(vt) for vt in list(valid_types)])}."
                f" Got {type(payload)}"
            )
        if isinstance(payload, (QalxEntity, uuid.UUID)):
            # Messages always get chunked and added in batches so make
            # sure the entity is a list if it is of these types
            payload = [payload]
        entries = []
        for _entity in payload:
            # Build our list of entries to ensure that they are all valid before
            # submitting to the queue
            entity_guid = _entity
            if isinstance(_entity, QalxEntity):
                # It's an instance of an entity - get the guid for submission
                # to the queue
                if not isinstance(_entity, (Item, Set, Group)):
                    raise QalxQueueError(
                        "Only Item, Set and Group entities can be submitted to the "
                        "queue. No messages submitted to queue."
                    )
                if _entity.entity_type != entity_type:
                    error = (
                        f"Mismatched entity types. Expected "
                        f"`{entity_type}`, got "
                        f"`{_entity.entity_type}`. Did you mean to do "
                        f"`{_entity.__class__.__name__}.add_to_queue("
                        f"entity, queue)?`. No messages submitted to queue."
                    )
                    raise QalxQueueError(error)
                entity_guid = _entity["guid"]

            entries.append(
                {
                    "Id": str(uuid.uuid4()),
                    "MessageBody": json.dumps(
                        {
                            "entity_type": entity_type,
                            # SQS needs the guid as a string - not UUID
                            "entity_guid": str(entity_guid),
                            **message_kwargs,
                        }
                    ),
                    # Unique MessageGroupId to prevent processing
                    # duplicates.
                    "MessageGroupId": str(uuid.uuid4()),
                    # Unique MessageDeduplicationId to allow the user
                    # to put the same entity on the queue multiple times
                    "MessageDeduplicationId": str(uuid.uuid4()),
                }
            )

        if entries:
            MAX_BATCH_SIZE = 10
            # Iterate through all entries and send them to the queue in batches
            for chunk in self._chunks(entries, chunk_size=MAX_BATCH_SIZE):
                chunked_objects = list(filter(None, chunk))
                responses.append(
                    self.broker_client.send_message_batch(
                        QueueUrl=self.info.queue_url, Entries=chunked_objects
                    )
                )

        return QueueResponse(responses)

    def purge(self):
        """
        Purge the queue. All messages on the queue will be deleted.

        This might take up to 60 seconds and any messages sent to the queue
        while it is being purged might also be deleted
        """
        self.broker_client.purge_queue(QueueUrl=self.info.queue_url)



import logging

import click
from click import ClickException, types

from pyqalx import QalxSession
from pyqalx.config import UserConfig, BotConfig
from pyqalx.core.errors import QalxError
from pyqalx.core.utils import _msg_user

from .. import cli_types
from ..cli.qalx import qalx
from ..utils import (
    TerminateBotTabulation,
    StopBotTabulation,
    ResumeBotTabulation,
    BotTabulation,
)


@qalx.command("bot-start")
@click.option(
    "-p",
    "--processes",
    type=cli_types.POSITIVE_INT,
    help="The number of workers to spawn in the bot.",
    default=1,
)
@click.option(
    "--skip-ini/--no-skip-ini",
    help="Should the bot use the profiles on disk?",
    default=False,
)
@click.argument("target", type=cli_types.BOT_IMPORT)
@click.option(
    "-q",
    "--queue-name",
    help="The name of the queue this bot should read messages from",
    required=True,
)
@click.option(
    "--qalx-session-class",
    type=cli_types.QALX_SESSION_IMPORT,
    help="The import path of a custom QalxSession class in the format "
    "`dotted.path.to:MyQalxSessionClass`",
    default=None,
)
@click.option(
    "--user-config-class",
    type=cli_types.USER_CONFIG_IMPORT,
    help="The import path of a custom UserConfigClass class in the format "
    "`dotted.path.to:MyUserConfigClass`",
    default=None,
)
@click.option(
    "--bot-config-class",
    type=cli_types.BOT_CONFIG_IMPORT,
    help="The import path of a custom BotConfigClass class in the format "
    "`dotted.path.to:MyBotConfigClass`",
    default=None,
)
@click.option(
    "--entity-class",
    type=cli_types.QALX_ENTITY_IMPORT,
    help="The import path of a custom QalxEntity class in the format "
    "`dotted.path.to:MyQalxEntityClass`",
    multiple=True,
)
@click.option(
    "--stop",
    is_flag=True,
    help="If provided the workers will be stopped immediately after startup",
)
@click.option(
    "--meta",
    type=cli_types.QALX_DICT,
    help=(
        "Key value pair of the format <key>=<value>, for example: "
        + "--meta <key>=<value>. Multiple meta arguments can be passed"
    ),
    multiple=True,
)
@click.option(
    "--workflow",
    type=click.STRING,
    help=(
        "The name of another bot to pass a message to once all step "
        + "functions have been completed on this bot"
    ),
    multiple=True,
)
@click.pass_context
def start(
    ctx,
    target,
    queue_name,
    processes,
    skip_ini,
    qalx_session_class,
    user_config_class,
    bot_config_class,
    entity_class,
    stop,
    meta,
    workflow,
):
    """
    Start the bot at TARGET

    This is the import path with a colon followed by the variable name
    of the bot. e.g. `my_qalx.my_bot:bot`
    """
    click.echo(
        f"Starting {target.name} reading from {queue_name} with"
        f" {processes} workers."
    )
    meta = {k: v for d in meta for k, v in d.items()} if meta else None
    user_profile = ctx.obj.get("USER_PROFILE", UserConfig.default)
    bot_profile = ctx.obj.get("BOT_PROFILE", BotConfig.default)
    qalx_session = QalxSession(
        profile_name=ctx.obj["USER_PROFILE"], skip_ini=skip_ini
    )
    try:
        target.start(
            queue_name=queue_name,
            user_profile_name=user_profile,
            bot_profile_name=bot_profile,
            processes=processes,
            skip_ini=skip_ini,
            qalx_session_class=qalx_session_class,
            user_config_class=user_config_class,
            bot_config_class=bot_config_class,
            entity_classes=list(entity_class),
            stop=stop,
            meta=meta,
            workflow=list(workflow),
        )
    except QalxError as exc:
        _msg_user(
            title="start",
            msg=str(exc),
            log=qalx_session.log,
            level=logging.ERROR,
        )
        _msg_user(msg="Bot Start Failed", log=qalx_session.log)


@qalx.command("bot-terminate")
@click.argument("name")
@click.pass_context
def terminate_bot(ctx, name):
    """Shutdown the bot named NAME"""
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", UserConfig.default)
    )
    bot_tabulation = TerminateBotTabulation(qalx_session, name)
    bot_to_kill = bot_tabulation.get_entity_or_display()
    if bot_to_kill:
        # Finally, terminate the bot
        try:
            qalx_session.bot.terminate(bot_to_kill)
        except QalxError as exc:
            _msg_user(
                title="terminate",
                msg=str(exc),
                log=qalx_session.log,
                level=logging.ERROR,
            )
            _msg_user(msg="Bot Terminate Failed", log=qalx_session.log)
        else:
            click.echo(
                f"Terminate signal sent for `{bot_to_kill.name} "
                f"({bot_to_kill.host.get('platform')})`.  Workers may take "
                f"some time to terminate"
            )


@qalx.command("bot-stop")
@click.argument("name")
@click.pass_context
def stop_bot(ctx, name):
    """Stop the bot named NAME.

    All workers to stop pulling jobs from the queue.
    """
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", UserConfig.default)
    )
    bot_tabulation = StopBotTabulation(qalx_session, name)
    bot_to_stop = bot_tabulation.get_entity_or_display()

    if bot_to_stop:
        try:
            qalx_session.bot.stop(bot_to_stop)
        except QalxError as exc:
            _msg_user(
                title="stop",
                msg=str(exc),
                log=qalx_session.log,
                level=logging.ERROR,
            )
            _msg_user(msg="Bot Stop Failed", log=qalx_session.log)
        else:
            click.echo(f"Stopping {name}.")


@qalx.command("bot-resume")
@click.argument("name")
@click.pass_context
def resume_bot(ctx, name):
    """Resume the bot named NAME.

    All workers to start pulling jobs from the queue.
    """
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", UserConfig.default)
    )
    bot_tabulation = ResumeBotTabulation(qalx_session, name)
    bot_to_resume = bot_tabulation.get_entity_or_display()
    if bot_to_resume:
        try:
            qalx_session.bot.resume(bot_to_resume)
        except QalxError as exc:
            _msg_user(
                title="resume",
                msg=str(exc),
                log=qalx_session.log,
                level=logging.ERROR,
            )
            _msg_user(msg="Bot Resume Failed", log=qalx_session.log)
        else:
            click.echo(f"Resuming {name}.")


@qalx.command("bot-info")
@click.argument("name")
@click.pass_context
def bot_info(ctx, name):
    """Print info about the bot named NAME."""
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", UserConfig.default)
    )
    try:
        bot_tabulation = BotTabulation(qalx_session, name, tabulate_single=True)
        bot_tabulation.get_entity_or_display()
    except QalxError as exc:
        _msg_user(
            title="info",
            msg=str(exc),
            log=qalx_session.log,
            level=logging.ERROR,
        )
        _msg_user(msg="Bot Info Failed", log=qalx_session.log)


@qalx.command("workers-terminate")
@click.argument("bot_name")
@click.argument("number", type=types.INT)
@click.pass_context
def terminate_workers(ctx, bot_name, number):
    """Terminate NUMBER of workers on bot named BOT_NAME."""
    qalx_session = QalxSession(
        profile_name=ctx.obj.get("USER_PROFILE", UserConfig.default)
    )
    bot_tabulation = TerminateBotTabulation(qalx_session, bot_name)
    bot_to_kill = bot_tabulation.get_entity_or_display()
    if bot_to_kill:
        if number > len(bot_to_kill.workers):
            raise ClickException(
                f"{number} is more than the number of workers on {bot_name}"
                f" ({len(bot_to_kill.workers)})"
            )
        for n in range(number):
            try:
                qalx_session.worker.terminate(
                    bot_to_kill.workers[n], bot_entity=bot_to_kill
                )
            except QalxError as exc:
                _msg_user(
                    title="workers terminate",
                    msg=str(exc),
                    log=qalx_session.log,
                    level=logging.ERROR,
                )
                _msg_user(msg="Worker Terminate Failed", log=qalx_session.log)
            else:
                click.echo(f"Terminated worker number {n + 1}.")



from collections import defaultdict

from pyqalx.core.errors import QalxFactoryBuildError
from pyqalx.factories.sectors.providers import _providers, LocalProvider


class Sectors(dict):
    """
    A dict subclass that groups the remote sectors into a dictionary
    of <provider_type>:<profile_name>:<region_name><sector_name>:<sector>
    This deep nesting will enable us to create a single stack for
    each provider:profile:region intersection

    Class attributes
        RESERVED_KEYS:  Specific keys at the top level of a stage that cannot
                        be used as the name of a sector.

    Instance attributes
        plan:           The factory plan as JSON
        stage:          The stage that the sectors are for
        factory:        The factory instance the sectors are built for
        validate_default_region:    Boolean of whether the default region
        should be validated
    """

    USER_PROFILE_KEY = "user_profile"
    BOT_PROFILE_KEY = "bot_profile"
    RESERVED_KEYS = {"workflow", USER_PROFILE_KEY, BOT_PROFILE_KEY}

    def __init__(
        self, factory, validate_default_region=True, validate_permissions=True
    ):
        """
        :param factory: The factory instance
        :param validate_default_region: Whether the default region should be
        validated.  The only time we don't want to ensure that there is a valid
        default region is when running the bootstrap script on remote bots as
        this uses the internal server to figure out the region
        :param validate_permissions: Whether the permissions of the calling
        user should be validated for remote sectors.  This won't need to happen
        when running the bootstrap script on remote bots as this uses the internal
        server to figure out the region
        """
        _sectors = {}
        for key in _providers.keys():
            _sectors[key] = defaultdict(lambda: defaultdict(dict))
        super(Sectors, self).__init__(_sectors)
        self.validate_default_region = validate_default_region
        self.validate_permissions = validate_permissions
        self.factory = factory
        self._group_sectors()

    @staticmethod
    def _get_provider(sector):
        sector_type = sector["type"]
        try:
            return _providers[sector_type]
        except KeyError:  # pragma: no cover
            # Should be handled by validation - this is here to
            # as a safety net.
            raise QalxFactoryBuildError(
                f"`{sector_type}` is not a valid "
                f"sector type.  Valid sector type"
                f' are `{", ".join(_providers.keys())}`'
            )

    def _group_sectors(self):
        """
        Groups the sectors on the plan by profile_name and then region_name.
        If no profile/region name supplied for a sector then use the default
        for the sectors provider.  Will ultimately lead to a group
        of sectors for each provider:profile_name:region_name intersection
        :return: A dict of sectors parameters grouped
         by profile_name:region_name
        """
        sectors = self._sectors()
        for sector_name, sector in sectors.items():
            provider = self._get_provider(sector)
            profile_name = sector[provider.PROFILE_NAME_KEY]
            region_name = sector[provider.REGION_NAME_KEY]

            self[provider._type][profile_name][region_name][
                sector_name
            ] = sector
        return self

    def _sectors(self):
        """
        Iterates over all the sectors on the stage and returns them grouped
        by profile_name/region_name.  All local sectors will be combined into
        a single sector.
        :return:The provider sectors as a dict of sector_name:sector_data
        """
        sectors = {}
        stage = self.factory._plan_stages[self.factory.stage]
        for sector_name, sector in stage.items():
            if sector_name not in self.RESERVED_KEYS:
                provider = self._get_provider(sector)
                if sector.get(provider.REGION_NAME_KEY) is None:
                    sector[
                        provider.REGION_NAME_KEY
                    ] = provider.default_region_name(
                        validate=self.validate_default_region,
                        log=self.factory.session.log,
                    )
                if provider._type == LocalProvider._type:
                    profile_name = provider.default_profile_name()
                else:
                    profile_name = self.factory.aws_profile
                sector[provider.PROFILE_NAME_KEY] = profile_name
                sector[self.USER_PROFILE_KEY] = self.factory.user_profile
                sector[self.BOT_PROFILE_KEY] = self.factory.bot_profile
                if self.validate_permissions:
                    profile_name = sector[provider.PROFILE_NAME_KEY]
                    region_name = sector[provider.REGION_NAME_KEY]
                    provider.validate_permissions(
                        profile_name=profile_name,
                        region_name=region_name,
                        log=self.factory.session.log,
                    )
                sectors[sector_name] = sector
        return sectors
